unit ActnRsyncU;

interface

uses
  MainConstants, SHUtilityProxy, SysUtils, StrUtils, Classes, ShlObj, Windows
  ;
type
  TActnRsync = class abstract(TObject)
  ///  Wrapper for rsync.exe
  ///  Copies files in an efficient manner, particularly to an rsync enabled backup server
  ///  Expects rsync.exe to be in the same folder as this application
  private
    class function RsyncFriendlyPath(const ASource: string): string;
    class function GetUserFolder: string;
  public
    class function Copy{To server}(const ASource, ADestination: string; ARecurse: boolean = true): string;
    class function Move{On server}(const ASource, ADestination: string): string;
    class function SSH{Command on server}(const ACommand: string): string;
    class procedure UploadKey(ALog: TStrings = nil);
    class procedure RunCommandSilent(const ACommand: string; ALog: TStrings = nil);
    class procedure RunCommandInteractive(const ACommand: string; ALog: TStrings = nil);
  end;

implementation

{ TActnRsync }

class function TActnRsync.Copy(const ASource, ADestination: string; ARecurse: boolean = true): string;
begin
  // Copy files from ASource on local computer to ADestination on rsync server.
  // ADestination is subfolder of BACKUPTO_FOLDER.
  // For now no checks on success.

  // Build command string

  // Basic command
  Result := 'rsync.exe';

  // Port
  if (IndexText(SERVER_PORT, ['', '22']) < 0) then
    // A non-standard port is specified
    Result := Result + ' -e "ssh -p ' + SERVER_PORT + '"';

  Result := Result + ' -a';
  // -a Archive option preserves file attributes including timestamp

  // -a turns on recursion.  Turn off if not needed
  if not ARecurse then
    Result := Result + ' --no-r';

  // Source
  Result := Result + ' ' + RsyncFriendlyPath(ASource);

  // Server
  // Colon on end indicated that destination is a remove server, not a local file name.
  // Use double colon for an rsync daemon.
  Result := Result + ' ' + SERVER_USERNAME + '@' + SERVER_URL + ':';

  // Destination on server
  if (ADestination <> '') then
    Result := Result + RsyncFriendlyPath(ADestination);
end;

class function TActnRsync.GetUserFolder: string;
///  Return the root of the users path using forward slashes, e.g.
///     c:\users\Simon
///  for the currently logged in user
var
  LPath: array[0..MAX_PATH] of WideChar;
  i: Integer;
begin
  // Use WINAPI function.
  // I have seen a post of the web that says CSIDL_PROFILE does not work for
  // administrator users but have not checked it
  SHGetFolderPath(0, CSIDL_PROFILE, 0, 0, LPath);

  Result := LPath;

  // Convert backslashes
  for i := 1 to Length(Result) do
    if (Result[i] = '\')then
      Result[i] := '/';
end;

class function TActnRsync.Move(const ASource, ADestination: string): string;
begin
  /// Build command string to
  /// Move or rename a file on a remote computer running SSH

  // Command that will run on server
  Result := Result + ' mv';

  // Source
  Result := Result + ' ' + ASource;

  // Destination
  Result := Result + ' ' + ADestination;

  // SSH prefix to call from client
  Result := SSH(Result);
end;

class function TActnRsync.RsyncFriendlyPath(const ASource: string): string;
var
  i: Integer;
begin
  // Returns the path in a manner which rsync.exe will accept

  Result := ASource;

  // all slashes must be forward
  for i := 1 to Length(Result) do
    if (Result[i] = '\') then
      Result[i] := '/';

  // Cannot have colon so odd handling for drive letter
  if (Length(Result) >= 2) and (Result[2] = ':') then
  begin
    // Must have "c:/foldername....." not "c:foldername....."
    if (Length(Result) < 3) or (Result[3] <> '/') then
      Insert('/', Result, 3);
    // Replace "c:" with "/cygdrive/c" which rsync requires
    Result := '/cygdrive/' + Result[1] + System.Copy(Result, 3, high(integer));
  end;

  Result := '"' + Result + '"';  // "" in case there are spaces in the folder/file names
end;

class procedure TActnRsync.RunCommandInteractive(const ACommand: string;
  ALog: TStrings);
var
  LCommand: string;
  LExitCode: cardinal;
begin
  ///  Run ACommand is a DOS box, leaving the box open, but do not return until
  ///  the box is closed.
  ///  Add the ACommand to ALog.
  // Note built command in log

  if (ALog <> nil) then
    ALog.Append(ACommand);

  // Run the command
  LCommand := '/k ' + ACommand;  // Because var parameter in following call
  SHCompleteProcess64('c:\Windows\System32\cmd.exe', LCommand, LExitCode, INFINITE, true);
end;

class procedure TActnRsync.RunCommandSilent(const ACommand: string;
  ALog: TStrings);
var
  LExitCode: cardinal;
begin
  ///  Run ACommand without any additional user display.
  ///  Add the ACommand to ALog.
  ///
  ///  This needs some error handling.
  ///  Error messages are not retreived.
  ///  If pass-keys are not shared so the hidden command prompts for a password
  ///  the Delphi application will lock up.

  // Note built command in log
  if (ALog <> nil) then
    ALog.Append(ACommand);

  // Run the command
  SHCompleteProcess(ACommand, LExitCode);
end;

class function TActnRsync.SSH(const ACommand: string): string;
begin
  // Return a string which will run the ACommand via SSH on the server

  // Basic command
  Result := 'ssh.exe';

  // Port
  if (IndexText(SERVER_PORT, ['', '22']) < 0) then
    // A non-standard port is specified
    Result := Result + ' -p ' + SERVER_PORT;

  // Server
  Result := Result + ' ' + SERVER_USERNAME + '@' + SERVER_URL;

  // Add the command
  Result := Result + ' ' + ACommand;
end;

class procedure TActnRsync.UploadKey(ALog: TStrings = nil);
///  Append file
///    [Client]c:/users/<username>/.ssh/id_rsa.pub
///  to
///    [Server]~/.ssh/authorized_keys
///
///  This code is for sharing key-pairs between cleint and server, so it is assumed
///  that the key pairs are not set up.  In which cases it will trigger interactive
///  prompts asking for the password.  For this reason it does not work.
///
///  Production code would need to
///  - Test for existance of file authorized_keys.  If it did not exist, then
///    rename id_rsa.pub to authorized_keys
///  - Delete the no-longer required id_rsa.pub
var
  sss:string;
begin
  //// Copy file to the server ////

  sss := Copy(
     RsyncFriendlyPath(GetUserFolder() + '/.ssh/id_rsa.pub')  // Source
   , '~/.ssh/'  // Destination
   , false);  // Not recusrive

  // Run the command interactively because there will be a prompt for the password
  RunCommandInteractive(sss, ALog);

  //// Rename file on the server ////

  // SSH concatenate command
  // ".copy" in filename is for testing purposes only
  sss := SSH('"cat  ~/.ssh/id_rsa.pub >> ~/.ssh/authorized_keys"');

  // Run the command interactively because there will be a prompt for the password
  RunCommandInteractive(sss, ALog);
end;

end.
